from typing import Protocol, TypeVar

from .job import Job

JobType = TypeVar("JobType", bound=Job, contravariant=True)


class StateStoreProtocol(Protocol[JobType]):
    def update(self, job: JobType) -> None: ...



import copy
from contextlib import ExitStack
from inspect import Parameter
from inspect import signature as get_signature
from typing import Annotated, Any, Callable, get_args, get_origin

from .annotations import Context, Depends, Input
from .depends_cache import DependsCache
from .types import Annotation


class FunctionArgumentInfo:
    def __init__(self) -> None:
        self._args_types: list[Annotation] = []
        self._kwarg_types: dict[str, Annotation] = {}

    def add_param(self, param: Parameter, type_: Annotation) -> None:
        if param.kind == Parameter.POSITIONAL_ONLY:
            self._args_types.append(type_)
            return
        self._kwarg_types[param.name] = type_

    def build_type(
        self,
        type_: Annotation,
        stack: ExitStack,
        depends_cache: DependsCache,
        user_context: Any,
        input_value: Any,
    ) -> Any:
        if isinstance(type_, Input):
            return input_value

        if isinstance(type_, Depends):
            return stack.enter_context(
                type_.create(
                    stack,
                    depends_cache,
                    user_context,
                    input_value,
                )
            )

        assert isinstance(type_, Context)
        return user_context

    def build_args(
        self,
        stack: ExitStack,
        depends_cache: DependsCache,
        user_context: Any,
        input_value: Any,
    ) -> tuple[list[Any], dict[str, Any]]:
        args: list[Any] = []
        kwargs: dict[str, Any] = {}

        for type_ in self._args_types:
            args.append(
                self.build_type(
                    type_,
                    stack,
                    depends_cache,
                    user_context,
                    input_value,
                )
            )

        for name, type_ in self._kwarg_types.items():
            kwargs[name] = self.build_type(
                type_,
                stack,
                depends_cache,
                user_context,
                input_value,
            )

        return args, kwargs


def get_param_info(param: Parameter) -> Input | Depends | Context:
    origin = get_origin(param.annotation)
    if origin is not Annotated:
        return Input()

    arguments = get_args(param.annotation)
    ergate_annotations = [
        argument
        for argument in arguments
        if isinstance(argument, (Input, Depends, Context))
    ]

    if not ergate_annotations:
        return Input()

    if len(ergate_annotations) > 1:
        raise ValueError(
            "Parameter annotations must contain no more than one dependency "
            f"or one context marker ({param.name=})"
        )

    return ergate_annotations[0]


def build_function_arg_info(function: Callable[..., Any]) -> FunctionArgumentInfo:
    signature = get_signature(function)
    function_wrapper = FunctionArgumentInfo()

    for param in signature.parameters.values():
        param_info = get_param_info(param)

        if isinstance(param_info, Depends):
            depends_copy = copy.copy(param_info)
            dependency_arg_info = build_function_arg_info(depends_copy.dependency)
            depends_copy.initialize(dependency_arg_info)
            param_info = depends_copy

        function_wrapper.add_param(param, param_info)

    return function_wrapper



from typing import Any

from pydantic import ValidationError  # noqa: F401


class ErgateError(Exception):
    """Base class for ergate exceptions."""


class InvalidDefinitionError(ErgateError):
    """Raised when a workflow/step definition is invalid."""


class AbortJob(ErgateError):  # noqa: N818
    """Raised from a step to abort a workflow.
    Should be interpreted as an expected failure.
    """

    def __init__(self, message: str) -> None:
        self.message = message


class SkipNSteps(ErgateError):  # noqa: N818
    """Raised from a step to skip N steps."""

    def __init__(self, n: int, retval: Any = None):
        self.n = n
        self.retval = retval



from typing import Any, Generic, TypeVar

from .job import Job
from .log import LOG
from .types import ExceptionHook

JobType = TypeVar("JobType", bound=Job)


class ErrorHookHandler(Generic[JobType]):
    def __init__(self) -> None:
        self._hooks: dict[Any, ExceptionHook[JobType]] = {}

    def register(
        self,
        exception: type[Exception],
        hook: ExceptionHook[JobType],
    ) -> None:
        self._hooks[exception] = hook

    def get_hook(self, exception: Exception) -> ExceptionHook[JobType] | None:
        for cls in type(exception).__mro__:
            if cls in self._hooks:
                return self._hooks[cls]
        return None

    def notify(self, job: JobType, exception: Exception) -> None:
        hook = self.get_hook(exception)
        if not hook:
            return

        try:
            hook(job, exception)
        except Exception:
            LOG.warning("Error hook raised an exception", exc_info=True)



from typing import Iterator

from .workflow import Workflow


class WorkflowRegistry:
    def __init__(self) -> None:
        self._workflows: dict[str, Workflow] = {}

    def __getitem__(self, unique_name: str) -> Workflow:
        try:
            return self._workflows[unique_name]
        except KeyError:
            raise KeyError(f'No workflow named "{unique_name}" is registered') from None

    def __iter__(self) -> Iterator[Workflow]:
        return iter(self._workflows.values())

    def register(self, workflow: Workflow) -> None:
        if workflow.unique_name in self._workflows:
            err = f'A workflow named "{workflow.unique_name}" is already registered'
            raise ValueError(err)
        self._workflows[workflow.unique_name] = workflow



from __future__ import annotations

from collections.abc import Generator
from contextlib import ExitStack, contextmanager
from typing import TYPE_CHECKING, Any, Callable, ParamSpec

from .depends_cache import DependsCache
from .inspect import build_function_arg_info

if TYPE_CHECKING:
    from .workflow import Workflow


CallableSpec = ParamSpec("CallableSpec")


class WorkflowStep:
    def __init__(
        self,
        workflow: Workflow,
        callable: Callable[CallableSpec, Any],
    ) -> None:
        self.workflow = workflow
        self.callable = callable
        self.arg_info = build_function_arg_info(callable)

    @property
    def name(self) -> str:
        return self.callable.__name__

    @contextmanager
    def build_args(
        self, user_context: Any, last_return_value: Any
    ) -> Generator[tuple[list[Any], dict[str, Any]], None, None]:
        with ExitStack() as stack:
            yield self.arg_info.build_args(
                stack,
                DependsCache(),
                user_context,
                last_return_value,
            )

    def __call__(self, *args: CallableSpec.args, **kwargs: CallableSpec.kwargs) -> Any:
        return self.callable(*args, **kwargs)

    def __str__(self) -> str:
        return f"{self.workflow.unique_name}.{self.callable.__name__}"



from typing import Generic, TypeVar

from .exceptions import AbortJob, SkipNSteps
from .handler import ErrorHookHandler
from .interrupt import DelayedKeyboardInterrupt
from .job import Job
from .log import LOG
from .queue import QueueProtocol
from .state_store import StateStoreProtocol
from .workflow_registry import WorkflowRegistry

JobType = TypeVar("JobType", bound=Job)


class JobRunner(Generic[JobType]):
    def __init__(
        self,
        queue: QueueProtocol[JobType],
        workflow_registry: WorkflowRegistry,
        state_store: StateStoreProtocol[JobType],
        error_hook_handler: ErrorHookHandler[JobType],
    ) -> None:
        self.queue = queue
        self.workflow_registry = workflow_registry
        self.state_store = state_store
        self.error_hook_handler = error_hook_handler

    def _run_job(self, job: JobType) -> None:
        input_value = job.get_input_value()

        workflow = self.workflow_registry[job.workflow_name]
        step_to_run = workflow[job.steps_completed]

        job.mark_running(step_to_run)
        self.state_store.update(job)

        try:
            LOG.info("Running %s - input value: %s", str(step_to_run), input_value)
            with step_to_run.build_args(job.user_context, input_value) as all_args:
                args, kwargs = all_args
                retval = step_to_run(*args, **kwargs)
        except AbortJob as exc:
            LOG.info("User requested to abort job: %s", exc)
            job.mark_aborted(exc.message)
        except SkipNSteps as exc:
            LOG.info("User requested to skip %d steps", exc.n)
            job.mark_n_steps_completed(exc.n + 1, exc.retval, len(workflow))
        except Exception as exc:
            LOG.exception("Job raised an exception")
            job.mark_failed(exc)
            self.error_hook_handler.notify(job, exc)
        else:
            LOG.info("Step completed successfully - return value: %s", retval)
            job.mark_n_steps_completed(1, retval, len(workflow))

        self.state_store.update(job)

        if job.should_be_requeued():
            LOG.info("Requeuing job")
            self.queue.put(job)

    def run(self) -> None:
        while True:
            LOG.info("Listening for next job")
            try:
                job = self.queue.get_one()
            except KeyboardInterrupt:
                return

            LOG.info("Job acquired")
            try:
                with DelayedKeyboardInterrupt():
                    self._run_job(job)
            except KeyboardInterrupt:
                return



from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field

from .job_status import JobStatus
from .workflow import WorkflowStep


class Job(BaseModel):
    id: Any = None
    workflow_name: str
    status: JobStatus = JobStatus.QUEUED
    steps_completed: int = Field(default=0, ge=0)
    percent_completed: float = Field(default=0.0, ge=0.0, le=100.0)
    initial_input_value: Any = None
    last_return_value: Any = None
    user_context: Any = None
    requested_start_time: datetime | None = None

    def get_input_value(self) -> Any:
        return (
            self.initial_input_value
            if self.steps_completed == 0
            else self.last_return_value
        )

    def mark_running(self, step: WorkflowStep) -> None:
        self.status = JobStatus.RUNNING

    def mark_failed(self, exception: Exception) -> None:
        self.status = JobStatus.FAILED

    def mark_n_steps_completed(
        self,
        n: int,
        return_value: Any,
        total_steps: int,
    ) -> None:
        self.steps_completed = min(self.steps_completed + n, total_steps)
        self.percent_completed = float((self.steps_completed / total_steps) * 100)
        self.status = (
            JobStatus.COMPLETED
            if self.steps_completed == total_steps
            else JobStatus.QUEUED
        )
        self.last_return_value = return_value

    def mark_aborted(self, message: str) -> None:
        self.status = JobStatus.ABORTED

    def should_be_requeued(self) -> bool:
        return self.status not in (
            JobStatus.ABORTED,
            JobStatus.COMPLETED,
            JobStatus.FAILED,
        )



import signal

from .log import LOG


class DelayedKeyboardInterrupt:
    def __enter__(self) -> None:
        self.signals_received: list[tuple] = []
        self.old_handlers = {
            signal.SIGTERM: signal.signal(signal.SIGTERM, self.handler),
            signal.SIGINT: signal.signal(signal.SIGINT, self.handler),
        }

    def handler(self, sig, frame) -> None:
        self.signals_received.append((sig, frame))
        signal_ = signal.Signals(sig)
        LOG.info("%s received - delaying it", signal_.name)

    def __exit__(self, type, value, traceback) -> None:
        for sig, old_handler in self.old_handlers.items():
            signal.signal(sig, old_handler)

        for signal_ in self.signals_received:
            old_handler = self.old_handlers[signal.Signals(signal_[0])]
            if not callable(old_handler):
                continue
            old_handler(*signal_)



from .annotations import Context, Depends, Input
from .app import Ergate
from .exceptions import (
    AbortJob,
    ErgateError,
    InvalidDefinitionError,
    SkipNSteps,
    ValidationError,
)
from .job import Job
from .job_status import JobStatus
from .queue import QueueProtocol
from .state_store import StateStoreProtocol
from .workflow import Workflow, WorkflowStep

__all__ = [
    "Context",
    "Depends",
    "Input",
    "Ergate",
    "ErgateError",
    "InvalidDefinitionError",
    "AbortJob",
    "SkipNSteps",
    "ValidationError",
    "Job",
    "StateStoreProtocol",
    "JobStatus",
    "QueueProtocol",
    "Workflow",
    "WorkflowStep",
]
